import re

from dagster_graphql.test.utils import execute_dagster_graphql, infer_pipeline_selector

from dagster import check
from dagster.core.execution.plan.objects import StepOutputHandle
from dagster.core.storage.intermediate_storage import build_fs_intermediate_storage
from dagster.utils import file_relative_path, merge_dicts
from dagster.utils.test import get_temp_file_name

from .graphql_context_test_suite import ReadonlyGraphQLContextTestMatrix
from .setup import (
    PoorMansDataFrame,
    csv_hello_world,
    csv_hello_world_solids_config,
    csv_hello_world_solids_config_fs_storage,
)

EXECUTION_PLAN_QUERY = """
query PipelineQuery($runConfigData: RunConfigData, $pipeline: PipelineSelector!, $mode: String!) {
  executionPlanOrError(runConfigData: $runConfigData, pipeline: $pipeline, mode: $mode) {
    __typename
    ... on ExecutionPlan {
      steps {
        key
        solidHandleID
        kind
        inputs {
          name
          type {
            name
          }
          dependsOn {
            key
          }
        }
        outputs {
          name
          type {
            name
          }
        }
        metadata {
          key
          value
        }
      }
    }
    ... on PipelineNotFoundError {
        pipelineName
    }
    ... on PipelineConfigValidationInvalid {
        pipelineName
        errors { message }
    }
    ... on PythonError {
        message
        stack
        cause {
            message
            stack
        }
    }
  }
}
"""

EXECUTE_PLAN_QUERY = """
mutation ($executionParams: ExecutionParams!, $retries: Retries) {
    executePlan(executionParams: $executionParams, retries: $retries) {
        __typename
        ... on ExecutePlanSuccess {
            pipeline { name }
            hasFailures
            stepEvents {
                __typename
                ... on MessageEvent {
                    message
                }
                stepKey
                ... on ExecutionStepOutputEvent {
                    outputName
                }
                ... on StepMaterializationEvent {
                    materialization {
                        label
                        description
                        metadataEntries {
                            ... on EventPathMetadataEntry {
                                label
                                description
                                path
                            }
                        }
                    }
                }
                ... on ExecutionStepFailureEvent {
                    error {
                        message
                    }
                }
            }
        }
        ... on PipelineConfigValidationInvalid {
            pipelineName
            errors { message }
        }
        ... on PipelineNotFoundError {
            pipelineName
        }
        ... on PythonError {
            message
            stack
            cause {
                message
                stack
            }
        }
    }
}
"""


def get_nameset(llist):
    return {item["name"] for item in llist}


def get_named_thing(llist, name):
    for cn in llist:
        if cn["name"] == name:
            return cn

    check.failed("not found")


# for snapshot testing remove any varying values
def clean_log_messages(result_data):
    for idx in range(len(result_data["executePlan"]["stepEvents"])):
        message = result_data["executePlan"]["stepEvents"][idx].get("message")
        if message is not None:
            result_data["executePlan"]["stepEvents"][idx]["message"] = re.sub(
                r"(\d+(\.\d+)?)", "{N}", message
            )
    return result_data


def test_success_whole_execution_plan(graphql_context, snapshot):
    run_config = csv_hello_world_solids_config_fs_storage()
    pipeline_run = graphql_context.instance.create_run_for_pipeline(
        pipeline_def=csv_hello_world, run_config=run_config
    )
    selector = infer_pipeline_selector(graphql_context, "csv_hello_world")
    result = execute_dagster_graphql(
        graphql_context,
        EXECUTE_PLAN_QUERY,
        variables={
            "executionParams": {
                "selector": selector,
                "runConfigData": run_config,
                "stepKeys": None,
                "executionMetadata": {"runId": pipeline_run.run_id},
                "mode": "default",
            },
        },
    )

    query_result = result.data["executePlan"]
    assert query_result["__typename"] == "ExecutePlanSuccess"
    assert query_result["pipeline"]["name"] == "csv_hello_world"
    assert query_result["hasFailures"] is False
    step_events = {
        step_event["stepKey"]: step_event
        for step_event in query_result["stepEvents"]
        if step_event["stepKey"]
    }
    assert "sum_solid.compute" in step_events
    assert "sum_sq_solid.compute" in step_events

    snapshot.assert_match(clean_log_messages(result.data))
    intermediate_storage = build_fs_intermediate_storage(
        graphql_context.instance.intermediates_directory, pipeline_run.run_id
    )
    assert intermediate_storage.has_intermediate(None, StepOutputHandle("sum_solid.compute"))
    assert intermediate_storage.has_intermediate(None, StepOutputHandle("sum_sq_solid.compute"))


def test_success_whole_execution_plan_with_filesystem_config(graphql_context, snapshot):
    instance = graphql_context.instance
    selector = infer_pipeline_selector(graphql_context, "csv_hello_world")
    run_config = merge_dicts(csv_hello_world_solids_config(), {"storage": {"filesystem": {}}})
    pipeline_run = instance.create_run_for_pipeline(
        pipeline_def=csv_hello_world, run_config=run_config
    )
    result = execute_dagster_graphql(
        graphql_context,
        EXECUTE_PLAN_QUERY,
        variables={
            "executionParams": {
                "selector": selector,
                "runConfigData": run_config,
                "stepKeys": None,
                "executionMetadata": {"runId": pipeline_run.run_id},
                "mode": "default",
            },
        },
    )

    query_result = result.data["executePlan"]

    assert query_result["__typename"] == "ExecutePlanSuccess"
    assert query_result["pipeline"]["name"] == "csv_hello_world"
    assert query_result["hasFailures"] is False
    step_events = {
        step_event["stepKey"]: step_event
        for step_event in query_result["stepEvents"]
        if step_event["stepKey"]
    }
    assert "sum_solid.compute" in step_events
    assert "sum_sq_solid.compute" in step_events

    snapshot.assert_match(clean_log_messages(result.data))
    intermediate_storage = build_fs_intermediate_storage(
        instance.intermediates_directory, pipeline_run.run_id
    )
    assert intermediate_storage.has_intermediate(None, StepOutputHandle("sum_solid.compute"))
    assert intermediate_storage.has_intermediate(None, StepOutputHandle("sum_sq_solid.compute"))


def test_success_whole_execution_plan_with_in_memory_config(graphql_context, snapshot):
    instance = graphql_context.instance
    selector = infer_pipeline_selector(graphql_context, "csv_hello_world")
    run_config = merge_dicts(csv_hello_world_solids_config(), {"storage": {"in_memory": {}}})
    pipeline_run = instance.create_run_for_pipeline(
        pipeline_def=csv_hello_world, run_config=run_config
    )
    result = execute_dagster_graphql(
        graphql_context,
        EXECUTE_PLAN_QUERY,
        variables={
            "executionParams": {
                "selector": selector,
                "runConfigData": run_config,
                "stepKeys": None,
                "executionMetadata": {"runId": pipeline_run.run_id},
                "mode": "default",
            },
        },
    )

    query_result = result.data["executePlan"]

    assert query_result["__typename"] == "ExecutePlanSuccess"
    assert query_result["pipeline"]["name"] == "csv_hello_world"
    assert query_result["hasFailures"] is False
    step_events = {
        step_event["stepKey"]: step_event
        for step_event in query_result["stepEvents"]
        if step_event["stepKey"]
    }
    assert "sum_solid.compute" in step_events
    assert "sum_sq_solid.compute" in step_events

    snapshot.assert_match(clean_log_messages(result.data))
    intermediate_storage = build_fs_intermediate_storage(
        instance.intermediates_directory, pipeline_run.run_id
    )
    assert not intermediate_storage.has_intermediate(None, StepOutputHandle("sum_solid.compute"))
    assert not intermediate_storage.has_intermediate(None, StepOutputHandle("sum_sq_solid.compute"))


def test_successful_one_part_execute_plan(graphql_context, snapshot):
    instance = graphql_context.instance
    run_config = csv_hello_world_solids_config_fs_storage()
    pipeline_run = instance.create_run_for_pipeline(
        pipeline_def=csv_hello_world, run_config=run_config
    )
    selector = infer_pipeline_selector(graphql_context, "csv_hello_world")

    result = execute_dagster_graphql(
        graphql_context,
        EXECUTE_PLAN_QUERY,
        variables={
            "executionParams": {
                "selector": selector,
                "runConfigData": run_config,
                "stepKeys": ["sum_solid.compute"],
                "executionMetadata": {"runId": pipeline_run.run_id},
                "mode": "default",
            },
        },
    )

    query_result = result.data["executePlan"]

    assert query_result["__typename"] == "ExecutePlanSuccess"
    assert query_result["pipeline"]["name"] == "csv_hello_world"
    assert query_result["hasFailures"] is False

    step_events = query_result["stepEvents"]

    assert [se["__typename"] for se in step_events] == [
        "ExecutionStepStartEvent",
        "ExecutionStepInputEvent",
        "ExecutionStepOutputEvent",
        "ObjectStoreOperationEvent",
        "ExecutionStepSuccessEvent",
    ]

    assert step_events[1]["stepKey"] == "sum_solid.compute"
    assert step_events[2]["outputName"] == "result"

    expected_value_repr = (
        """[OrderedDict([('num1', '1'), ('num2', '2'), ('sum', 3)]), """
        """OrderedDict([('num1', '3'), ('num2', '4'), ('sum', 7)])]"""
    )

    assert step_events[3]["stepKey"] == "sum_solid.compute"
    assert step_events[4]["stepKey"] == "sum_solid.compute"

    snapshot.assert_match(clean_log_messages(result.data))

    intermediate_storage = build_fs_intermediate_storage(
        instance.intermediates_directory, pipeline_run.run_id
    )
    assert intermediate_storage.has_intermediate(None, StepOutputHandle("sum_solid.compute"))
    assert (
        str(
            intermediate_storage.get_intermediate(
                None, PoorMansDataFrame, StepOutputHandle("sum_solid.compute")
            ).obj
        )
        == expected_value_repr
    )


def test_successful_two_part_execute_plan(graphql_context, snapshot):
    instance = graphql_context.instance
    run_config = csv_hello_world_solids_config_fs_storage()
    pipeline_run = instance.create_run_for_pipeline(
        pipeline_def=csv_hello_world, run_config=run_config
    )
    selector = infer_pipeline_selector(graphql_context, "csv_hello_world")
    result_one = execute_dagster_graphql(
        graphql_context,
        EXECUTE_PLAN_QUERY,
        variables={
            "executionParams": {
                "selector": selector,
                "runConfigData": run_config,
                "stepKeys": ["sum_solid.compute"],
                "executionMetadata": {"runId": pipeline_run.run_id},
                "mode": "default",
            },
        },
    )

    assert result_one.data["executePlan"]["__typename"] == "ExecutePlanSuccess"

    snapshot.assert_match(clean_log_messages(result_one.data))

    result_two = execute_dagster_graphql(
        graphql_context,
        EXECUTE_PLAN_QUERY,
        variables={
            "executionParams": {
                "selector": selector,
                "runConfigData": csv_hello_world_solids_config_fs_storage(),
                "stepKeys": ["sum_sq_solid.compute"],
                "executionMetadata": {"runId": pipeline_run.run_id},
                "mode": "default",
            },
        },
    )

    query_result = result_two.data["executePlan"]
    assert query_result["__typename"] == "ExecutePlanSuccess"
    assert query_result["pipeline"]["name"] == "csv_hello_world"
    assert query_result["hasFailures"] is False
    step_events = query_result["stepEvents"]
    assert [se["__typename"] for se in step_events] == [
        "ExecutionStepStartEvent",
        "ObjectStoreOperationEvent",
        "ExecutionStepInputEvent",
        "ExecutionStepOutputEvent",
        "ObjectStoreOperationEvent",
        "ExecutionStepSuccessEvent",
    ]
    assert step_events[0]["stepKey"] == "sum_sq_solid.compute"
    assert step_events[1]["stepKey"] == "sum_sq_solid.compute"
    assert step_events[2]["stepKey"] == "sum_sq_solid.compute"
    assert step_events[3]["outputName"] == "result"
    assert step_events[4]["stepKey"] == "sum_sq_solid.compute"

    snapshot.assert_match(clean_log_messages(result_two.data))

    expected_value_repr = (
        """[OrderedDict([('num1', '1'), ('num2', '2'), ('sum', 3), """
        """('sum_sq', 9)]), OrderedDict([('num1', '3'), ('num2', '4'), ('sum', 7), """
        """('sum_sq', 49)])]"""
    )

    intermediate_storage = build_fs_intermediate_storage(
        instance.intermediates_directory, pipeline_run.run_id
    )

    assert intermediate_storage.has_intermediate(None, StepOutputHandle("sum_sq_solid.compute"))
    assert (
        str(
            intermediate_storage.get_intermediate(
                None, PoorMansDataFrame, StepOutputHandle("sum_sq_solid.compute")
            ).obj
        )
        == expected_value_repr
    )


class TestExecutionPlan(ReadonlyGraphQLContextTestMatrix):
    def test_invalid_config_fetch_execute_plan(self, graphql_context, snapshot):
        selector = infer_pipeline_selector(graphql_context, "csv_hello_world")
        result = execute_dagster_graphql(
            graphql_context,
            EXECUTION_PLAN_QUERY,
            variables={
                "pipeline": selector,
                "runConfigData": {
                    "solids": {"sum_solid": {"inputs": {"num": {"csv": {"path": 384938439}}}}}
                },
                "mode": "default",
            },
        )

        assert not result.errors
        assert result.data
        assert (
            result.data["executionPlanOrError"]["__typename"] == "PipelineConfigValidationInvalid"
        )
        assert len(result.data["executionPlanOrError"]["errors"]) == 1
        assert (
            "Invalid scalar at path root:solids:sum_solid:inputs:num"
            in result.data["executionPlanOrError"]["errors"][0]["message"]
        )
        result.data["executionPlanOrError"]["errors"][0][
            "message"
        ] = "Invalid scalar at path root:solids:sum_solid:inputs:num"
        snapshot.assert_match(result.data)


def test_invalid_config_execute_plan(graphql_context, snapshot):
    selector = infer_pipeline_selector(graphql_context, "csv_hello_world")
    result = execute_dagster_graphql(
        graphql_context,
        EXECUTE_PLAN_QUERY,
        variables={
            "executionParams": {
                "selector": selector,
                "runConfigData": {
                    "solids": {"sum_solid": {"inputs": {"num": {"csv": {"path": 384938439}}}}}
                },
                "stepKeys": [
                    "sum_solid.num.input_thunk",
                    "sum_solid.compute",
                    "sum_sq_solid.compute",
                ],
                "executionMetadata": {"runId": "kdjkfjdfd"},
                "mode": "default",
            },
        },
    )

    assert not result.errors
    assert result.data
    assert result.data["executePlan"]["__typename"] == "PipelineConfigValidationInvalid"
    assert len(result.data["executePlan"]["errors"]) == 1
    assert (
        "Invalid scalar at path root:solids:sum_solid:inputs:num"
        in result.data["executePlan"]["errors"][0]["message"]
    )
    result.data["executePlan"]["errors"][0][
        "message"
    ] = "Invalid scalar at path root:solids:sum_solid:inputs:num"
    snapshot.assert_match(result.data)


def test_pipeline_not_found_error_execute_plan(graphql_context, snapshot):
    selector = infer_pipeline_selector(graphql_context, "nope")
    result = execute_dagster_graphql(
        graphql_context,
        EXECUTE_PLAN_QUERY,
        variables={
            "executionParams": {
                "selector": selector,
                "runConfigData": {
                    "solids": {"sum_solid": {"inputs": {"num": {"csv": {"path": "ok"}}}}}
                },
                "stepKeys": [
                    "sum_solid.num.input_thunk",
                    "sum_solid.compute",
                    "sum_sq_solid.compute",
                ],
                "executionMetadata": {"runId": "kdjkfjdfd"},
                "mode": "default",
            },
        },
    )

    assert result.data["executePlan"]["__typename"] == "PipelineNotFoundError"
    assert result.data["executePlan"]["pipelineName"] == "nope"
    snapshot.assert_match(result.data)


def test_basic_execute_plan_with_materialization(graphql_context):
    selector = infer_pipeline_selector(graphql_context, "csv_hello_world")
    with get_temp_file_name() as out_csv_path:

        run_config = {
            "solids": {
                "sum_solid": {
                    "inputs": {"num": file_relative_path(__file__, "../data/num.csv")},
                    "outputs": [{"result": out_csv_path}],
                }
            }
        }

        result = execute_dagster_graphql(
            graphql_context,
            EXECUTION_PLAN_QUERY,
            variables={"pipeline": selector, "runConfigData": run_config, "mode": "default",},
        )

        steps_data = result.data["executionPlanOrError"]["steps"]

        assert set([step_data["key"] for step_data in steps_data]) == set(
            ["sum_solid.compute", "sum_sq_solid.compute",]
        )

        instance = graphql_context.instance

        pipeline_run = instance.create_run_for_pipeline(
            pipeline_def=csv_hello_world, run_config=run_config
        )

        result = execute_dagster_graphql(
            graphql_context,
            EXECUTE_PLAN_QUERY,
            variables={
                "executionParams": {
                    "selector": selector,
                    "runConfigData": run_config,
                    "stepKeys": ["sum_solid.compute", "sum_sq_solid.compute"],
                    "executionMetadata": {"runId": pipeline_run.run_id},
                    "mode": "default",
                },
            },
        )

        assert result.data

        step_mat_event = None

        for message in result.data["executePlan"]["stepEvents"]:
            if message["__typename"] == "StepMaterializationEvent":
                # ensure only one event
                assert step_mat_event is None
                step_mat_event = message

        # ensure only one event
        assert step_mat_event
        assert step_mat_event["materialization"]
        assert len(step_mat_event["materialization"]["metadataEntries"]) == 1
        metadata_entry = step_mat_event["materialization"]["metadataEntries"][0]
        assert metadata_entry["path"] == out_csv_path



from dagster_graphql.client.mutations import (
    execute_execute_plan_mutation,
    execute_execute_plan_mutation_raw,
)

from dagster import file_relative_path
from dagster.cli.workspace.cli_target import PythonFileTarget, workspace_from_load_target
from dagster.core.definitions.reconstructable import (
    ReconstructablePipeline,
    get_ephemeral_repository_name,
)
from dagster.core.instance import DagsterInstance

EXPECTED_EVENTS = {
    ("STEP_INPUT", "sleeper.compute"),
    ("STEP_INPUT", "sleeper_2.compute"),
    ("STEP_INPUT", "sleeper_3.compute"),
    ("STEP_INPUT", "sleeper_4.compute"),
    ("STEP_INPUT", "total.compute"),
    ("STEP_OUTPUT", "giver.compute"),
    ("STEP_OUTPUT", "sleeper.compute"),
    ("STEP_OUTPUT", "sleeper_2.compute"),
    ("STEP_OUTPUT", "sleeper_3.compute"),
    ("STEP_OUTPUT", "sleeper_4.compute"),
    ("STEP_OUTPUT", "total.compute"),
    ("STEP_START", "giver.compute"),
    ("STEP_START", "sleeper.compute"),
    ("STEP_START", "sleeper_2.compute"),
    ("STEP_START", "sleeper_3.compute"),
    ("STEP_START", "sleeper_4.compute"),
    ("STEP_START", "total.compute"),
    ("STEP_SUCCESS", "giver.compute"),
    ("STEP_SUCCESS", "sleeper.compute"),
    ("STEP_SUCCESS", "sleeper_2.compute"),
    ("STEP_SUCCESS", "sleeper_3.compute"),
    ("STEP_SUCCESS", "sleeper_4.compute"),
    ("STEP_SUCCESS", "total.compute"),
}


def load_sleepy_workspace(instance):
    return workspace_from_load_target(
        PythonFileTarget(
            file_relative_path(__file__, "sleepy.py"), "sleepy_pipeline", working_directory=None
        ),
        instance,
    )


def sleepy_recon_pipeline():
    return ReconstructablePipeline.for_file(
        file_relative_path(__file__, "sleepy.py"), "sleepy_pipeline"
    )


def test_execute_execute_plan_mutation():
    pipeline_name = "sleepy_pipeline"
    instance = DagsterInstance.local_temp()

    pipeline = sleepy_recon_pipeline()
    workspace = load_sleepy_workspace(instance)
    pipeline_run = instance.create_run_for_pipeline(pipeline_def=pipeline.get_definition())
    variables = {
        "executionParams": {
            "runConfigData": {},
            "mode": "default",
            "selector": {
                "repositoryLocationName": get_ephemeral_repository_name(pipeline_name),
                "repositoryName": get_ephemeral_repository_name(pipeline_name),
                "pipelineName": pipeline_name,
            },
            "executionMetadata": {"runId": pipeline_run.run_id},
        }
    }
    result = execute_execute_plan_mutation(workspace, variables, instance_ref=instance.get_ref())
    seen_events = set()
    for event in result:
        seen_events.add((event.event_type_value, event.step_key))

    assert seen_events == EXPECTED_EVENTS


def test_execute_execute_plan_mutation_raw():
    pipeline_name = "sleepy_pipeline"
    pipeline = sleepy_recon_pipeline()
    instance = DagsterInstance.local_temp()

    workspace = load_sleepy_workspace(instance)

    pipeline_run = instance.create_run_for_pipeline(pipeline_def=pipeline.get_definition())
    variables = {
        "executionParams": {
            "runConfigData": {},
            "mode": "default",
            "selector": {
                "repositoryLocationName": get_ephemeral_repository_name(pipeline_name),
                "repositoryName": get_ephemeral_repository_name(pipeline_name),
                "pipelineName": pipeline_name,
            },
            "executionMetadata": {"runId": pipeline_run.run_id},
        }
    }
    result = execute_execute_plan_mutation_raw(
        workspace, variables, instance_ref=instance.get_ref()
    )
    seen_events = set()
    for event in result:
        seen_events.add((event.dagster_event.event_type_value, event.step_key))

    assert seen_events == EXPECTED_EVENTS



import sys

from dagster.core.host_representation import PythonEnvRepositoryLocation, RepositoryLocationHandle
from dagster.core.types.loadable_target_origin import LoadableTargetOrigin
from dagster.utils import file_relative_path


def test_dagster_out_of_process_location():
    env = PythonEnvRepositoryLocation(
        RepositoryLocationHandle.create_python_env_location(
            location_name="test_location",
            loadable_target_origin=LoadableTargetOrigin(
                executable_path=sys.executable,
                python_file=file_relative_path(__file__, "setup.py"),
                attribute="test_repo",
            ),
        )
    )
    assert env.get_repository("test_repo")



from dagster_graphql.test.utils import execute_dagster_graphql

from dagster import repository
from dagster.core.host_representation import external_repository_data_from_def
from dagster.seven import mock

from .graphql_context_test_suite import GraphQLContextVariant, make_graphql_context_test_suite

# import dagster.api.utils


class TestReloadRepositoriesOutOfProcess(
    make_graphql_context_test_suite(
        context_variants=[GraphQLContextVariant.readonly_in_memory_instance_out_of_process_env()]
    )
):
    def test_out_of_process_reload_location(self, graphql_context):
        result = execute_dagster_graphql(
            graphql_context, RELOAD_REPOSITORY_LOCATION_QUERY, {"repositoryLocationName": "test"}
        )

        assert result
        assert result.data
        assert result.data["reloadRepositoryLocation"]
        assert result.data["reloadRepositoryLocation"]["__typename"] == "RepositoryLocation"
        assert result.data["reloadRepositoryLocation"]["name"] == "test"
        assert result.data["reloadRepositoryLocation"]["repositories"] == [{"name": "test_repo"}]
        assert result.data["reloadRepositoryLocation"]["isReloadSupported"] is True

        with mock.patch(
            # note it where the function is *used* that needs to mocked, not
            # where it is defined.
            # see https://docs.python.org/3/library/unittest.mock.html#where-to-patch
            "dagster.api.snapshot_repository.execute_unary_api_cli_command"
        ) as cli_command_mock:

            @repository
            def new_repo():
                return []

            new_repo_data = external_repository_data_from_def(new_repo)

            cli_command_mock.return_value = new_repo_data

            result = execute_dagster_graphql(
                graphql_context,
                RELOAD_REPOSITORY_LOCATION_QUERY,
                {"repositoryLocationName": "test"},
            )

            assert cli_command_mock.call_count == 1

            assert result.data["reloadRepositoryLocation"]["repositories"] == [{"name": "new_repo"}]


class TestReloadRepositoriesInProcess(
    make_graphql_context_test_suite(
        context_variants=[GraphQLContextVariant.readonly_in_memory_instance_in_process_env()]
    )
):
    def test_reload_not_supported(self, graphql_context):
        result = execute_dagster_graphql(
            graphql_context,
            RELOAD_REPOSITORY_LOCATION_QUERY,
            {"repositoryLocationName": "<<in_process>>"},
        )
        assert result.data["reloadRepositoryLocation"]["__typename"] == "ReloadNotSupported"

    def test_location_not_found(self, graphql_context):
        result = execute_dagster_graphql(
            graphql_context, RELOAD_REPOSITORY_LOCATION_QUERY, {"repositoryLocationName": "nope"},
        )
        assert result.data["reloadRepositoryLocation"]["__typename"] == "RepositoryLocationNotFound"


RELOAD_REPOSITORY_LOCATION_QUERY = """
mutation ($repositoryLocationName: String!) {
   reloadRepositoryLocation(repositoryLocationName: $repositoryLocationName) {
      __typename
      ... on RepositoryLocation {
        name
        repositories {
            name
        }
        isReloadSupported
      }
   }
}
"""


class TestReloadRepositoriesManagedGrpc(
    make_graphql_context_test_suite(
        context_variants=[GraphQLContextVariant.readonly_in_memory_instance_managed_grpc_env(),]
    )
):
    def test_managed_grpc_reload_location(self, graphql_context):
        result = execute_dagster_graphql(
            graphql_context, RELOAD_REPOSITORY_LOCATION_QUERY, {"repositoryLocationName": "test"}
        )

        assert result
        assert result.data
        assert result.data["reloadRepositoryLocation"]
        assert result.data["reloadRepositoryLocation"]["__typename"] == "RepositoryLocation"
        assert result.data["reloadRepositoryLocation"]["name"] == "test"
        assert result.data["reloadRepositoryLocation"]["repositories"] == [{"name": "test_repo"}]
        assert result.data["reloadRepositoryLocation"]["isReloadSupported"] is True



from dagster_graphql.cli import execute_query

from dagster.core.errors import DagsterError
from dagster.core.events.log import DagsterEventRecord
from dagster.core.instance import DagsterInstance
from dagster.serdes import deserialize_json_to_dagster_namedtuple

from .query import EXECUTE_PLAN_MUTATION, RAW_EXECUTE_PLAN_MUTATION
from .util import HANDLED_EVENTS, dagster_event_from_dict


class DagsterGraphQLClientError(DagsterError):
    """Indicates that some error has occurred when executing a GraphQL query against
    dagster-graphql"""


def execute_execute_plan_mutation(workspace, variables, instance_ref=None):
    with (
        DagsterInstance.from_ref(instance_ref) if instance_ref else DagsterInstance.ephemeral()
    ) as instance:
        res = execute_query(
            workspace, EXECUTE_PLAN_MUTATION, variables, use_sync_executor=True, instance=instance
        )
    handle_execution_errors(res, "executePlan")
    return handle_execute_plan_result(res)


def handle_execution_errors(res, expected_type):
    if res is None:
        raise DagsterGraphQLClientError("Unhandled error type. Missing query response")

    if res.get("errors"):
        raise DagsterGraphQLClientError(
            "Internal error in GraphQL request. Response: {}".format(res)
        )

    if not res.get("data", {}).get(expected_type, {}).get("__typename"):
        raise DagsterGraphQLClientError("Unexpected response type. Response: {}".format(res))


def execute_execute_plan_mutation_raw(recon_repo, variables, instance_ref=None):
    """The underlying mutation returns the DagsterEventRecords serialized as strings, rather
    than dict representations of the DagsterEvents, thus "raw". This method in turn returns a
    stream of DagsterEventRecords, not DagsterEvents."""

    with (
        DagsterInstance.from_ref(instance_ref) if instance_ref else DagsterInstance.ephemeral()
    ) as instance:
        res = execute_query(recon_repo, RAW_EXECUTE_PLAN_MUTATION, variables, instance=instance)
        handle_execution_errors(res, "executePlan")
        return handle_execute_plan_result_raw(res)


def handle_error_states(res_type, res_data):
    if res_type == "InvalidStepError":
        raise DagsterGraphQLClientError(
            "invalid step {step_key}".format(step_key=res_data["invalidStepKey"])
        )

    if res_type == "PipelineConfigValidationInvalid":
        errors = [err["message"] for err in res_data["errors"]]
        raise DagsterGraphQLClientError(
            "Pipeline configuration invalid:\n{errors}".format(errors="\n".join(errors))
        )

    if res_type == "PipelineNotFoundError":
        raise DagsterGraphQLClientError(
            'Pipeline "{pipeline_name}" not found: {message}:'.format(
                pipeline_name=res_data["pipelineName"], message=res_data["message"]
            )
        )

    if res_type == "PythonError":
        raise DagsterGraphQLClientError(
            "Subplan execution failed: {message}\n{stack}".format(
                message=res_data["message"], stack=res_data["stack"]
            )
        )


def handle_execute_plan_result(res):
    res_data = res["data"]["executePlan"]

    res_type = res_data["__typename"]

    handle_error_states(res_type, res_data)

    if res_type == "ExecutePlanSuccess":
        pipeline_name = res_data["pipeline"]["name"]

        return [
            dagster_event_from_dict(e, pipeline_name)
            for e in res_data["stepEvents"]
            if e["__typename"] in HANDLED_EVENTS
        ]

    raise DagsterGraphQLClientError("Unexpected result type")


def handle_execute_plan_result_raw(res):
    res_data = res["data"]["executePlan"]

    res_type = res_data["__typename"]

    handle_error_states(res_type, res_data)

    if res_type == "ExecutePlanSuccess":
        raw_event_records = [
            DagsterEventRecord(
                event_record.error_info,
                event_record.message,
                event_record.level,
                event_record.user_message,
                event_record.run_id,
                event_record.timestamp,
                event_record.step_key,
                event_record.pipeline_name,
                event_record.dagster_event,
            )
            for event_record in [
                deserialize_json_to_dagster_namedtuple(e) for e in res_data["rawEventRecords"]
            ]
        ]
        return raw_event_records

    raise DagsterGraphQLClientError("Unexpected result type")
